<?php

namespace Drupal\send_emails;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Url;
// For vertification.
use Drupal\user\UserInterface;
/* Classes used in the __construct */
use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\Mail\MailManager;
use Drupal\Core\Config\ConfigFactory;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\Datetime\DateFormatter;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Render\Renderer;
use Drupal\Core\Session\AccountInterface;
use Psr\Container\ContainerInterface;
use Symfony\Component\HttpFoundation\RequestStack;
use Drupal\auto_login_url\AutoLoginUrlCreate;

/**
 * Provides settings for notify user.
 *
 * @package Drupal\EmailApi
 */
class EmailApi {
  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManager
   */
  protected $entityTypeManager;

  /**
   * The mail manager.
   *
   * @var \Drupal\Core\Mail\MailManager
   */
  protected $mailManager;

  /**
   * The connection to database.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $connection;

  /**
   * The date formatter.
   *
   * @var \Drupal\Core\Datetime\DateFormatter
   */
  protected $dateFormatter;

  /**
   * Config Email.
   *
   * @var \Drupal\Core\Config\ConfigFactory
   */
  private $emailConfig;

  /**
   * The logger channel factory.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  private $logger;

  /**
   * Store the manually set body template.
   *
   * @var string
   */
  private $bodyTemplateOverride;

  /**
   * Store the manually set subject template .
   *
   * @var string
   */
  private $subjectTemplateOverride;

  /**
   * The time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $timeService;

  /**
   * The messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * Current User.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * Renderer service.
   *
   * @var Drupal\Core\Render\Renderer
   */
  protected $renderer;

  /**
   * Auto Login Service Creator
   * 
   * @var Drupal\auto_login_url\AutoLoginUrlCreate
   */
  protected $aluService;

  /**
   * Class constructor.
   */
  public function __construct(EntityTypeManager $entity_type_manager,
  MailManager $mail_manager,
  ConfigFactory $config,
  LoggerChannelFactoryInterface $logger,
  Connection $connection,
  DateFormatter $dateFormatter,
  TimeInterface $timeService,
  MessengerInterface $messenger,
  RequestStack $requestStack,
  AccountInterface $currentUser,
  Renderer $renderer,
  AutoLoginUrlCreate $autoLoginUrlCreate
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->mailManager = $mail_manager;
    $this->connection = $connection;
    $this->dateFormatter = $dateFormatter;
    $this->messenger = $messenger;

    // $this->logger = $logger;
    $this->logger = $logger->get('Send Emails');

    // For Email and Site Name.
    $this->siteConfig = $config->get('system.site');

    $this->emailConfig = $config->get('send_emails.settings');

    // The Auto Login Service from the auto_login_url module.
    $this->aluService = \Drupal::service('auto_login_url.create');

    // Miscellenous.
    $this->timeService = $timeService;
    $this->requestStack = $requestStack;
    $this->currentUser = $currentUser;
    $this->renderer = $renderer;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.mail'),
      $container->get('config.factory'),
      $container->get('logger.factory'),
      $container->get('database'),
      $container->get('date.formatter'),
      $container->get('datetime.time'),
      $container->get('messenger'),
      $container->get('request_stack'),
      $container->get('current_user'),
      $container->get('renderer'),
      $container->get('auto_login_url.create')
    );
  }

  /**
   * Notify active users with a role.
   *
   * @param string $roleToNotify
   *   String the machine name of the role users must have to receive an email.
   * @param string $template
   *   String the template to send emails to.
   * @param string $destination
   *   The path to redirect to after logging in.
   * @param string $replyTo
   *   The email to send emails when replying.
   */
  public function notifyUsersByRole(string $roleToNotify, string $template, string $destination = '', $replyTo = '') {

    $users = $this->entityTypeManager
      ->getStorage('user')
      ->loadByProperties(
              [
                'status' => 1,
                'roles' => $roleToNotify,
              ]
          );

    return $this->notifyByUsers($users, $template, $destination, $replyTo);
  }

  /**
   * Notify Users by providing an array of users.
   *
   * @param array $users
   *   UserInterface[] the users to send emails to.
   * @param string $template
   *   String the template to send emails to.
   * @param string $destination
   *   The path to redirect to after logging in.
   * @param string $replyTo
   *   The email to send emails when replying.
   */
  public function notifyByUsers(array $users, string $template, string $destination = '', $replyTo = '') {
    $result = TRUE;

    foreach ($users as $user) {
      $tries = 0;
      $emailResult = FALSE;

      while ($tries < 5 && !$emailResult) {
        $emailResult = $this->notifyUser($user, $template, $destination, $replyTo);
        $tries++;
      }

      if (!$emailResult) {
        $result = FALSE;
        $this->messenger->addError(
              $this->t(
                  'Unable to notify %user &lt;%email&gt; after @num tries',
                  [
                    '%user' => $user->getAccountName(),
                    '%email' => $user->getEmail(),
                    '@num' => $tries,
                  ]
              )
          );
      }
    }

    return $result;
  }

  /**
   * Notify Users by providing an array of users.
   *
   * @param array $emailDetails
   *   The information to send emails to['name' => '', 'email' => ''].
   * @param string $template
   *   String the template to send emails to.
   * @param string $destination
   *   The path to redirect to after logging in.
   * @param string $replyTo
   *   The email to send emails when replying.
   * @param array $data
   *   Data that can be accessed in the TWIG template.
   */
  public function notifyByDetails(array $emailDetails, string $template, string $destination = '', $replyTo = '', array $data = []) {
    $result = TRUE;

    $dataErrors = [];
    $emailErrors = [];

    foreach ($emailDetails as $emailDetail) {
      if (!isset($emailDetail['name'], $emailDetail['email'])) {
        $result = FALSE;
        $dataErrors[] = $emailDetail;
      }
      else {
        $tries = 0;
        $emailResult = FALSE;

        while ($tries < 5 && !$emailResult) {
          $emailResult = $this->notifyEmail($emailDetail, $template, $destination, $replyTo, $data);
          $tries++;
        }

        if (!$emailResult) {
          $result = FALSE;
          $emailErrors[] = $emailDetail['name'] . ' <' . $emailDetail['email'] . '>';
        }
      }

    }

    if (!empty($dataErrors)) {
      $this->logger->error(
            $this->t(
                'Malformed data when sending %template: @data',
                ['%template' => $template, '@data' => json_encode($dataErrors)]
            )
        );
    }

    if (!empty($emailErrors)) {
      $this->messenger->addError(
            $this->t(
                'Unable to notify %count users: %userList',
                [
                  '%count' => count($emailErrors),
                  '%userList' => implode(',', $emailErrors),
                ]
            )
        );
      $this->logger->error(
            $this->t(
                'Failed emails when sending %template: @data',
                ['%template' => $template, '@data' => json_encode($emailErrors)]
            )
        );
    }

    return $result;
  }

  /**
   * Public function to notify a single user by entity or id.
   *
   * @param int $user
   *   UserInterface|int the user entity or user id to send emails to.
   * @param string $template
   *   String the template to send emails to.
   * @param string $destination
   *   The path to redirect to after logging in.
   * @param string $replyTo
   *   The email to send emails when replying.
   * @param array $data
   *   Data that can be accessed in the TWIG template.
   * @param string $toEmail
   *   The email to send to.
   *
   * @return bool
   *   TRUE if the email was sent, FALSE if not.
   */
  public function notifyUser($user, string $template, string $destination = '', string $replyTo = '', array $data = [], string $toEmail = '') {

    // Load User.
    if (!($user instanceof UserInterface)) {
      if (is_int($user)) {
        $userStorage = $this->entityTypeManager->getStorage('user');
        $user = $userStorage->load($user);
      }
      else {
        $this->logger->error('User is not loaded properly');
        return FALSE;
      }
    }

    // Auto login link to the destination page
    // Remove first slash so URL is local, so trusted.
    $destinationTrimmed = ltrim($destination, '/');
    $autoLoginLink = $this->aluService->create($user->id(), $destinationTrimmed, TRUE);

    // Prepare Body Data.
    $data = [
      'auto_login_link' => $autoLoginLink,
      'misc' => [
        'userEntity' => $user,
      ],
    ] + $data;

    // Prepare Email.
    $userData = [
      'name' => $user->getDisplayName(),
      'email' => $toEmail === '' ? $user->getEmail() : $toEmail,
    ];

    if (empty($userData['email'])) {
      $this->messenger->addWarning(
            $this->t(
                'Unable to notify <a href="@userLink">%user</a> because the user does not have an email',
                [
                  '%user' => $user->getDisplayName(),
                  '@userLink' => $user->toUrl('edit-form')->toString(),
                ]
            )
        );
      return FALSE;
    }
    else {
      $returnResult = $this->notifyEmail($userData, $template, $destination, $replyTo, $data);
    }

    return $returnResult;
  }

  /**
   * Public function to notify using a name and email.
   *
   * @param array $userData
   *   The information to send emails to ['name' => '', 'email' => ''].
   * @param string $template
   *   String the template to send emails to.
   * @param string $destination
   *   The path to redirect to after logging in.
   * @param string $replyTo
   *   The email to send emails when replying.
   * @param array $data
   *   Data that can be accessed in the TWIG template.
   *
   * @throws \Exception
   *   When $userData is not properly formatted.
   *
   * @return bool|array
   *   Result if the email was sent, FALSE if not.
   */
  public function notifyEmail(array $userData, string $template, string $destination = '', string $replyTo = '', array $data = []) {

    if (!isset($userData['name'], $userData['email'])) {
      throw new \Exception('User data does not contain "name" or "email" key.');
    }

    // Set default data.
    if (empty($destination)) {
      $destination = $this->emailConfig->get('emails.' . $template . '.destination') ?? '/';
    }

    if (empty($replyTo)) {
      $replyTo = $this->emailConfig->get('emails.' . $template . '.replyTo') ?? '';
    }

    // Prepare link to destination.
    if (substr($destination, 0, 1) === '/') {
      $destination = $this->requestStack->getSchemeAndHttpHost() . $destination;
    }

    // Prepare Body Data.
    $data['name'] = $userData['name'];
    $data['destination_link'] = $destination;
    $data['misc']['time-raw'] = $this->timeService->getRequestTime();

    // Prepare Email.
    $to = $userData['email'];
    $returnResult = $this->sendTemplate($to, $template, $data, $replyTo);

    return $returnResult;
  }

  /**
   * Private function that returns the headers for the emails.
   *
   * @return bool|array
   *   Email was sent, FALSE if not.
   */
  public function sendSimple($to, $subject, $body) {
    $replyTo = $this->siteConfig->get('mail');
    $message['to'] = $to;
    $message['subject'] = $subject;
    $message['body'][] = $body;
    $message['headers'] = $this->getHeaders($replyTo);
    $message['params'] = [
      'html' => TRUE,
    ];
    $message['send'] = TRUE;

    $mailer = new PhpMail();
    $message = $mailer->format($message);

    return $mailer->mail($message);
  }

  /**
   * Function to send template email through.
   */
  public function sendTemplateEmail(string $to, string $template, $data = [], $replyTo = NULL, $send = TRUE) {

    return $this->sendTemplate($to, $template, $data, $replyTo, $send);
  }

  /**
   * Public function to send emails based on emails.
   *
   * @return bool
   *   TRUE if email template is available,FALSE if template is not available.
   */
  private function sendTemplate(string $to, string $template, $data = [], $replyTo = '', $send = TRUE) {

    if (empty($this->emailConfig->get('emails.' . $template))) {
      $this->logger->error(
            'Email Template @template does not exist', [
              '@template' => $template,
            ]
        );
      return FALSE;
    }

    // Set reply to site email by default.
    if (empty($replyTo)) {
      $replyTo = $this->siteConfig->get('mail');
    }

    // Add additional data.
    $data['site_name'] = $data['site_name'] ?? $this->siteConfig->get('name');
    $data['site_front'] = Url::fromRoute('<front>');

    // Format subject & body.
    $subject = $this->getSubjectFromTemplate($template, $data);
    $body = $this->getBodyFromTemplate($template, $data);

    // Set the message.
    $message['to'] = $to;
    $message['subject'] = $subject;
    $message['body'][] = $body;
    $message['headers'] = $this->getHeaders($replyTo);
    $message['params'] = [
      'html' => TRUE,
    ];

    $module = 'send_emails';
    $key = 'send_emails.template.' . $template;
    $to = $to;
    $params['message'] = $message;
    $langcode = $this->currentUser->getPreferredLangcode();
    $send = TRUE;

    $result = $this->mailManager->mail($module, $key, $to, $langcode, $params, NULL, $send);

    return $result;

  }

  /**
   * Allow the body template to be override.
   *
   * @todo Refactor to reduce the number of parameters passed down functions
   */
  public function overrideBodyTemplate(string $bodyTemplateOverride) {
    $this->bodyTemplateOverride = $bodyTemplateOverride;
  }

  /**
   * Format the Template & allow for overrides.
   */
  private function getBodyFromTemplate(string $template, array $data) {
    $templateText = '';
    if (!empty($this->bodyTemplateOverride)) {
      $templateText = $this->bodyTemplateOverride;
    }
    else {
      $templateText = $this->emailConfig->get('emails.' . $template . '.body');
    }

    $templateText = '<body style="font-size: 14px; color: #000;">' . $templateText . '</body>';
    $render = [
      '#type' => 'inline_template',
      '#template' => $templateText,
      '#context' => $data,
    ];

    return $this->renderer->renderPlain($render);
    ;
  }

  /**
   * Allow the subject template to be override.
   *
   * @todo Refactor to reduce the number of parameters passed down functions
   */
  public function overrideSubjectTemplate(string $subjectTemplateOverride) {
    $this->subjectTemplateOverride = $subjectTemplateOverride;
  }

  /**
   * Format the Template & allow for overrides.
   */
  private function getSubjectFromTemplate(string $template, array $data) {
    $templateText = '';
    if (!empty($this->subjectTemplateOverride)) {
      $templateText = $this->subjectTemplateOverride;
    }
    else {
      $templateText = $this->emailConfig->get('emails.' . $template . '.subject');
    }

    $render = [
      '#type' => 'inline_template',
      '#template' => $templateText,
      '#context' => $data,
    ];

    return $this->renderer->renderPlain($render);
    ;
  }

  /**
   * Private function that returns the headers for the emails.
   *
   * @return array
   *   The headers for the email.
   */
  private function getHeaders($replyTo) {
    // $hostName = \Drupal::request()->getHost();
    // if (self::endsWith($replyTo, '@' . $hostName)) {
    // $from = $replyTo;
    // }
    // else {
    // $from = $this->siteConfig->get('mail');
    // }
    // Can not override manually.
    $from = $this->siteConfig->get('mail');
    $headers = [
      'content-type' => 'text/html',
      'MIME-Version' => '1.0',
      'reply-to' => $replyTo,
      'from' => $this->siteConfig->get('name') . ' <' . $from . '>',
    ];

    return $headers;
  }

}
